import os

import numpy as np
import pytest

from chemicalspace.layers.diversity import (
    DIVERSITY_METHODS,
    diversity_methods_dict,
    ChemicalSpaceDiversityLayer,
)

np.random.seed(42)

TESTS_DIR = os.path.dirname(os.path.abspath(__file__))

INPUT_SMI_FILE = os.path.join(TESTS_DIR, "data", "inputs1.smi")


@pytest.fixture
def space() -> ChemicalSpaceDiversityLayer:
    return ChemicalSpaceDiversityLayer.from_smi(INPUT_SMI_FILE)


def test_uniqueness(space: ChemicalSpaceDiversityLayer) -> None:
    uniqueness = space.uniqueness()
    assert isinstance(uniqueness, float)
    assert 0 <= uniqueness <= 1


@pytest.mark.parametrize("method", tuple(diversity_methods_dict.keys()))
def test_diversity(
    space: ChemicalSpaceDiversityLayer, method: DIVERSITY_METHODS
) -> None:
    diversity = space.diversity(method=method)
    assert isinstance(diversity, float)
    assert 0 <= diversity <= 1



import os

import numpy as np
import pytest

from chemicalspace.layers.acquisition import STRATEGIES, ChemicalSpaceAcquisitionLayer

np.random.seed(42)

TESTS_DIR = os.path.dirname(os.path.abspath(__file__))

INPUT_SMI_FILE = os.path.join(TESTS_DIR, "data", "inputs1.smi")


@pytest.fixture
def space() -> ChemicalSpaceAcquisitionLayer:
    cs = ChemicalSpaceAcquisitionLayer.from_smi(INPUT_SMI_FILE)
    cs.scores = np.array(range(len(cs)), dtype=int)  # assign scores
    return cs


@pytest.mark.parametrize("strategy", ["random", "maxmin"])
@pytest.mark.parametrize("n", (1, 5, 10))
def test_pick_noscores(
    space: ChemicalSpaceAcquisitionLayer, strategy: STRATEGIES, n: int
) -> None:
    space_picked: ChemicalSpaceAcquisitionLayer = space.pick(n=n, strategy=strategy)

    assert isinstance(space_picked, ChemicalSpaceAcquisitionLayer)
    assert len(space_picked) == n
    assert len(space - space_picked) == len(space) - n


@pytest.mark.parametrize("strategy", ["greedy"])
@pytest.mark.parametrize("n", (1, 5, 10))
def test_pick_scores(
    space: ChemicalSpaceAcquisitionLayer, strategy: STRATEGIES, n: int
) -> None:
    space_picked: ChemicalSpaceAcquisitionLayer = space.pick(n=n, strategy=strategy)

    assert isinstance(space_picked, ChemicalSpaceAcquisitionLayer)
    assert len(space_picked) == n
    assert len(space - space_picked) == len(space) - n



from typing import Optional, Callable, Dict

import numpy as np
import scipy
from numpy.typing import NDArray
from sklearn.metrics import pairwise_distances_chunked, pairwise_distances
from typing_extensions import Literal, TypeAlias

from chemicalspace.utils import ArrayIntOrFloat, parallel_map
from .base import ChemicalSpaceBaseLayer


def reduce_sum(x, _):
    return x.sum(axis=1)


def internal_distance(
    X: ArrayIntOrFloat,
    metric: str,
    n_jobs: Optional[int] = None,
    working_memory: Optional[int] = None,
) -> float:
    """
    Compute the average pairwise distance between all points in `X`,
    given a distance metric.

    The calculation is done in chunks, and parallelized.

    See `sklearn.metrics.pairwise_distances_chunked` for more details.

    Args:
        X (Union[NDArray[int], NDArray[float]]): The input data.
        metric (str): The distance metric to use.
        n_jobs (Optional[int], optional): The number of jobs to run in parallel.
            Defaults to None.
        working_memory (Optional[int], optional): The amount of memory to use for
            the computation. Defaults to None.

    Returns:
        float: The average pairwise distance between all points in `X`.

    """

    n = X.shape[0] ** 2
    total = 0

    for chunk in pairwise_distances_chunked(
        X, metric=metric, n_jobs=n_jobs, working_memory=working_memory
    ):
        total += chunk.sum()  # type: ignore

    return total / n


def vendi_score(
    X: ArrayIntOrFloat,
    metric: str,
    n_jobs: Optional[int] = None,
) -> float:
    """
    Compute the Vendi score for an array of points,
    normalized for the number of data points.

    The Vendi score is a diversity evaluation metric for machine learning models.
    It is based on the entropy of the eigenvalues of the pairwise distance matrix.

    From Friedman et al.
    "The Vendi Score: A Diversity Evaluation Metric for Machine Learning"
    https://arxiv.org/abs/2210.02410

    Args:
        X (Union[NDArray[int], NDArray[float]]): The input data.
        metric (str): The distance metric to use.
        n_jobs (Optional[int], optional): The number of jobs to run in parallel.
            Defaults to None.

    Returns:
        float: The Vendi score for the input data.

    """

    pairwise = pairwise_distances(X, metric=metric, n_jobs=n_jobs)
    pairwise_scaled = pairwise / X.shape[0]

    evals: NDArray[np.float_] = scipy.linalg.eigvalsh(pairwise_scaled)
    evals_nonzero = evals[evals > 0]

    entropy = -(evals_nonzero * np.log(evals_nonzero)).sum()
    score = np.exp(entropy)

    score_per_point = score / X.shape[0]

    return score_per_point


DIVERSITY_METHODS: TypeAlias = Literal["internal-distance", "vendi"]
diversity_methods_dict: Dict[str, Callable] = {
    "internal-distance": internal_distance,
    "vendi": vendi_score,
}


class ChemicalSpaceDiversityLayer(ChemicalSpaceBaseLayer):

    def diversity(
        self, method: DIVERSITY_METHODS = "internal-distance", **kwargs
    ) -> float:
        """
        Compute the diversity of the chemical space data.

        Args:
            method (DIVERSITY_METHODS): The diversity algorithm to use.
                Defaults to 'internal-distance'.
            **kwargs: Additional keyword arguments to pass to the diversity algorithm.

        Returns:
            float: The diversity of the chemical space data.

        """

        if method not in diversity_methods_dict:
            raise ValueError(
                f"Unknown diversity method: {method}. "
                f"Allowed methods: {list(diversity_methods_dict.keys())}"
            )

        return diversity_methods_dict[method](self.features, self.metric, **kwargs)

    def uniqueness(self) -> float:
        """
        Return the fraction of unique molecules in the chemical space.

        Returns:
            float: The fraction of unique molecules in the chemical space.
        """

        n_unique = len(set(parallel_map(self.hash_mol, self.mols)))
        return n_unique / len(self)



import os

import pytest

from chemicalspace import ChemicalSpace

TESTS_DIR = os.path.dirname(os.path.abspath(__file__))

INPUT_SMI_FILE = os.path.join(TESTS_DIR, "data", "inputs1.smi")


@pytest.fixture
def space() -> ChemicalSpace:
    return ChemicalSpace.from_smi(INPUT_SMI_FILE)


# Here we just test that ChemicalSpace is a subclass of the right layers
def test_attributes(space):

    assert hasattr(space, "mols")
    assert hasattr(space, "indices")
    assert hasattr(space, "scores")
    assert space._features is None

    # Base
    assert hasattr(space, "chunks")
    # Clustering
    assert hasattr(space, "cluster")
    # Neighbors
    assert hasattr(space, "carve")
    # Projection
    assert hasattr(space, "project")
    # Acquisition
    assert hasattr(space, "pick")



import warnings
from abc import ABC, abstractmethod
from functools import lru_cache, partial
from typing import (
    Any,
    Dict,
    Generator,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    get_args,
)

import numpy as np
from numpy.typing import NDArray
from rdkit.Chem import Mol  # type: ignore
from typing_extensions import TypeAlias

from chemicalspace.utils import SEED, hash_mol

from .base import ChemicalSpaceBaseLayer, T

ClusteringMethodsType: TypeAlias = Literal[
    "kmedoids", "agglomerative-clustering", "sphere-exclusion", "scaffold"
]
CLUSTERING_METHODS: Tuple[str, ...] = get_args(ClusteringMethodsType)

ClusteringMethodsTypeN: TypeAlias = Literal[
    "kmedoids", "agglomerative-clustering"
]  # methods that require n_clusters
CLUSTERING_METHODS_N: Tuple[str, ...] = get_args(ClusteringMethodsTypeN)


class BaseClusteringX(ABC):
    """
    BaseClusteringMethod is an abstract class that defines the
    interface for clustering methods that take in an array of features and return an
    array of cluster labels.
    """

    @abstractmethod
    def fit_predict(self, X: NDArray[Any]) -> NDArray[np.int_]:
        """
        Perform clustering on the input data.

        Args:
            X (ndarray): The input data to cluster.

        Returns:
            NDArray[int]: An array of cluster labels for each point.
        """
        raise NotImplementedError


class BaseClusteringMols(ABC):
    """
    BaseClusteringMols is an abstract class that defines the
    interface for clustering methods that take in an array of RDKit molecules and
    return an array of cluster labels.
    """

    @abstractmethod
    def fit_predict(self, mols: Sequence[Mol], **kwargs) -> NDArray[np.int_]:
        """
        Perform clustering on the input data.

        Args:
            mols (Sequence[Mol]): The input molecules to cluster.
            **kwargs: Additional keyword arguments.

        Returns:
            NDArray[int]: An array of cluster labels for each molecule.
        """
        raise NotImplementedError


def get_optimal_cluster_number(
    features: NDArray[Any],
    model: Any,
    min_clusters: int = 2,
    max_clusters: Optional[int] = None,
    metric: str = "jaccard",
    n_runs: int = 10,
) -> int:
    """
    Calculates the optimal number of clusters using the silhouette score.

    Args:
        features (ndarray): The input features for clustering.
        model (Any): The clustering model to use. Must be instantiable,
            have a `n_clusters` parameter, and a `fit_predict` method.
        min_clusters (int, optional): The minimum number of clusters to consider.
            Defaults to 2.
        max_clusters (int, optional): The maximum number of clusters to consider.
        metric (str, optional): The metric to use for silhouette score.
            Defaults to "jaccard".
        n_runs (int, optional): The number of runs to average the silhouette score.
            Defaults to 10.

    Returns:
        int: The optimal number of clusters.

    """
    from sklearn.metrics import silhouette_score

    if max_clusters is None:
        max_clusters = len(features) // 2

    scan_n = np.linspace(min_clusters, max_clusters, n_runs, dtype=int)
    scan_n = np.unique(scan_n)

    scores: List[float] = []
    for n_clusters in scan_n:
        model_inst = model(n_clusters=n_clusters)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            labels = model_inst.fit_predict(features)
            score = silhouette_score(features, labels, metric=metric)
        scores.append(float(score))

    n = int(scan_n[np.argmax(scores)])

    return n


class SphereExclusion(BaseClusteringX):
    """
    A class representing a clustering algorithm based on sphere exclusion.
    It clusters input points in a set of clusters such that the minimum
    distance between any two clusters is (approximately) greater than a given radius.
    """

    def __init__(self, radius: float = 0.4, metric: str = "jaccard", **kwargs):
        """
        Initialize the SphereExclusion clustering algorithm.

        Args:
            radius (float, optional): The radius of the sphere. Defaults to 0.4.
            metric (str, optional): The metric to use for clustering.
                Defaults to "jaccard".
            **kwargs: Implemented for compatibility. Ignored.
        """
        self.radius = radius
        self.metric = metric
        _ = kwargs  # discard

    def fit_predict(self, X: NDArray[Any]) -> NDArray[np.int_]:
        """
        Perform clustering on the input data.

        Args:
            X (ndarray): The input data to cluster.

        Returns:
            NDArray[int]: An array of cluster labels for each point.
        """

        from sklearn.neighbors import BallTree

        tree = BallTree(X, metric=self.metric)

        cluster_idx = 0

        labels = np.full(X.shape[0], fill_value=-1, dtype=int)

        for i in range(X.shape[0]):
            if labels[i] != -1:
                continue

            idx = tree.query_radius(
                X[i : i + 1],
                r=self.radius,
                return_distance=False,
            )[0]

            neighbor_labels: Set[int] = set(labels[idx]) - {-1}
            for label in neighbor_labels:
                mask = labels == label
                labels[mask] = cluster_idx

            labels[idx] = cluster_idx
            cluster_idx += 1

        # Renumber the clusters to start from 0
        unique_labels = np.unique(labels)
        for i, label in enumerate(unique_labels):
            labels[labels == label] = i

        return labels


class ScaffoldClustering(BaseClusteringMols):
    """
    A class representing a clustering algorithm based on Murcko scaffolds.
    """

    def __init__(self, generic: bool = True, **kwargs):
        """
        Initialize the ScaffoldClustering algorithm.

        Args:
            generic (bool, optional): Whether to use generic scaffolds. Defaults to True
            **kwargs: Implemented for compatibility. Ignored
        """
        self.generic = generic
        _ = kwargs

    def fit_predict(
        self, mols: Union[Sequence[Mol], NDArray[Mol]], **kwargs  # type: ignore
    ) -> NDArray[np.int_]:
        """
        Perform clustering on the input data.

        Args:
            mols (Sequence[Mol] | NDArray[Mol]): The input molecules to cluster.
            **kwargs: Additional keyword arguments. Ignored.

        Returns:
            NDArray[int]: Cluster labels

        """
        from rdkit.Chem.Scaffolds import MurckoScaffold  # type: ignore

        _ = kwargs  # discard

        labels: List[int] = []
        cluster_mapping: Dict[str, int] = {}
        c = 0
        for mol in mols:
            murcko = MurckoScaffold.GetScaffoldForMol(mol)
            if self.generic:
                murcko = MurckoScaffold.MakeScaffoldGeneric(murcko)

            if murcko.GetNumAtoms() > 0:
                murcko_hash = hash_mol(murcko)
            else:
                murcko_hash = ""

            if murcko_hash in cluster_mapping:
                murcko_label = cluster_mapping[murcko_hash]
            else:
                cluster_mapping[murcko_hash] = c
                murcko_label = c
                c += 1

            labels.append(murcko_label)

        return np.array(labels)


# @dataclass(frozen=False, repr=False)
class ChemicalSpaceClusteringLayer(ChemicalSpaceBaseLayer):
    """
    A class representing a layer for clustering chemical space data.

    Inherits from ChemicalSpaceBaseLayer.

    Methods:
        cluster: Perform clustering on the chemical space data.
        ksplits: Generate train-test splits based on clustering.

    Attributes:
        Inherits attributes from ChemicalSpaceBaseLayer.
    """

    @lru_cache
    def cluster(
        self,
        n_clusters: Optional[int] = None,
        method: ClusteringMethodsType = "kmedoids",
        seed: int = SEED,
        **kwargs,
    ) -> NDArray[np.int_]:
        """
        Perform clustering on the chemical space data.

        Args:
            n_clusters (int | None): The number of clusters to create
                (Ignored if not used). If None, the number of clusters will be
                determined by silhouette score
            method (str): The clustering method to use.
            seed (int, optional): The random seed for reproducibility. Defaults to SEED.
            **kwargs: Additional keyword arguments to pass to the clustering algorithm.

        Returns:
            NDArray[int]: An array of cluster labels for each molecule.

        """
        if method == "kmedoids":
            from sklearn_extra.cluster import KMedoids

            obj = partial(KMedoids, metric=self.metric, random_state=seed)

            if n_clusters is None:
                n_clusters = get_optimal_cluster_number(self.features, obj, **kwargs)

        elif method == "agglomerative-clustering":
            from sklearn.cluster import AgglomerativeClustering

            obj = partial(
                AgglomerativeClustering, metric=self.metric, linkage="complete"
            )

            if n_clusters is None:
                n_clusters = get_optimal_cluster_number(self.features, obj, **kwargs)

        elif method == "sphere-exclusion":
            obj = partial(SphereExclusion, metric=self.metric, **kwargs)
            n_clusters = -1

        elif method == "scaffold":
            obj = partial(ScaffoldClustering, **kwargs)
            n_clusters = -1

        else:
            raise ValueError(f"Invalid clustering method: {method}")

        clusterer = obj(n_clusters=n_clusters, **kwargs)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if isinstance(clusterer, BaseClusteringMols):
                labels = np.array(clusterer.fit_predict(self.mols), dtype=int)
            else:
                labels = np.array(clusterer.fit_predict(self.features), dtype=int)

        n_clusters_actual = len(set(labels))

        if n_clusters is not None:
            if (n_clusters > 0) and (n_clusters_actual != n_clusters):
                raise ValueError(
                    f"Number of clusters ({len(set(labels))}) does not "
                    f"match number of clusters requested ({n_clusters})."
                )

        return labels

    def yield_clusters(
        self,
        n_clusters: Optional[int] = None,
        method: ClusteringMethodsType = "kmedoids",
        seed: int = SEED,
        **kwargs,
    ) -> Generator[T, Any, None]:  # type: ignore
        """
        Yields clusters from the chemical space.

        Args:
            n_clusters (int | None): The number of clusters to create
                (Ignored if not used). If None, the number of clusters will be
                determined by silhouette score
            method (CLUSTERING_METHODS): The clustering method to use
            seed (int, optional): The random seed for reproducibility. Defaults to SEED
            **kwargs: Additional keyword arguments to be passed to the clustering method

        Yields:
            Generator[ChemicalSpaceBaseLayer, Any, None]: A generator that yields each
                cluster as a ChemicalSpaceBaseLayer object.
        """
        labels = self.cluster(
            n_clusters=n_clusters,
            method=method,
            seed=seed,
            **kwargs,
        )

        n_cluster_actual = len(set(labels))

        if (n_clusters is not None) and (n_cluster_actual != n_clusters):
            raise ValueError(
                f"Number of clusters ({n_cluster_actual}) does not match number "
                f"of clusters requested ({n_clusters})."
            )

        for i in range(n_cluster_actual):
            mask = np.array(labels == i)
            yield self[mask]

    def split(
        self,
        n_splits: Optional[int] = None,
        method: ClusteringMethodsType = "kmedoids",
        seed: int = SEED,
        **kwargs,
    ) -> Generator[Tuple[T, T], None, None]:
        """
        Generate train-test splits based on clustering.

        Args:
            n_splits (Optional[int]): The number of splits to generate,
                if supported by the clustering method.
            method (str): The clustering method to use.
            seed (int, optional): The random seed for reproducibility. Defaults to SEED.
            **kwargs: Additional keyword arguments to pass to the clustering algorithm.

        Yields:
            A tuple of ChemicalSpaceBaseLayer objects representing
                the train and test splits.

        """
        labels = self.cluster(
            n_clusters=n_splits,
            method=method,
            seed=seed,
            **kwargs,
        )

        n_cluster_actual = len(set(labels))

        if (n_splits is not None) and (n_cluster_actual != n_splits):
            raise ValueError(
                f"Number of clusters ({n_cluster_actual}) does not match number of "
                f"clusters requested ({n_splits})."
            )

        n_splits_actual = len(set(labels))

        for i in range(n_splits_actual):
            mask = np.array(labels == i)
            train: T = self[~mask]
            test: T = self[mask]
            yield train, test



import warnings
from abc import ABC
from functools import cached_property
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

import numpy as np
from numpy.typing import NDArray
from rdkit import Chem
from rdkit.Chem import Mol  # type: ignore
from typing_extensions import TypeAlias

from chemicalspace.utils import (
    MaybeIndex,
    MaybeScore,
    MolFeaturizerType,
    MolOrSmiles,
    Number,
    SliceType,
    factory,
    parallel_map,
)

from .. import utils

T = TypeVar("T", bound="ChemicalSpaceBaseLayer")
ScoreArray: TypeAlias = Union[NDArray[np.int_], NDArray[np.float_], NDArray[np.bool_]]


class ChemicalSpaceBaseLayer(ABC):
    """
    The base class for the ChemicalSpace object.
    Implements the core functionality for the ChemicalSpace object.
    """

    def __init__(
        self,
        mols: Sequence[MolOrSmiles],
        indices: Optional[Sequence[Any]] = None,
        scores: Optional[Sequence[Number]] = None,
        featurizer: MolFeaturizerType = utils.ecfp4_featurizer,
        metric: str = "jaccard",
        features: Optional[NDArray[Any]] = None,
        hash_indices: bool = False,
        n_jobs: int = 1,
    ) -> None:
        """
        Initializes a ChemicalSpace object.

        Args:
            mols (Sequence[Mol | str]):
                A sequence of RDKit Mol objectsr SMILES strings.
            indices (Optional[Sequence[Any]], optional):
                A sequence of indices for the molecules. Defaults to None.
            scores (Optional[Sequence[Number]], optional):
                A sequence of scores for the molecules. Defaults to None.
            featurizer (MolFeaturizerType, optional):
                The featurizer to use for the molecules.
                Defaults to utils.ecfp4_featurizer.
            metric (str, optional):
                The sklearn/scipy metric to use for the featurizer.
                Defaults to "jaccard".
            features (Optional[NDArray[Any]], optional):
                Precomputed features for the molecules. Defaults to None.
            hash_indices (bool, optional):
                Whether to include indices in the hash. Defaults to False.
            n_jobs (int, optional): The number of parallel jobs to run. Defaults to 1.


        Raises:
            ValueError: If the number of indices does not match the number of molecules.
            ValueError: If the number of scores does not match the number of molecules.
        """
        mols_m = np.array((parallel_map(utils.safe_smiles2mol, mols, n_jobs=n_jobs)))
        self.mols: NDArray[Mol] = np.array(mols_m)  # type: ignore

        self.indices: Union[NDArray[Any], None] = (
            np.array(indices) if indices is not None else None
        )
        self.scores: Union[ScoreArray, None] = (
            np.array(scores) if scores is not None else None
        )
        self.n_jobs = n_jobs
        self.featurizer = featurizer
        self.metric = metric
        self.hash_indices = hash_indices

        if self.indices is not None and len(self.indices) != len(self.mols):
            raise ValueError("Number of indices must match number of molecules")
        if self.scores is not None and len(self.scores) != len(self.mols):
            raise ValueError("Number of scores must match number of molecules")
        if (features is not None) and (len(features) != len(self.mols)):
            raise ValueError("Number of features must match number of molecules")

        self._features: Union[NDArray[Any], None] = features

        self.name = self.__class__.__name__

    def add(
        self, mol: MolOrSmiles, idx: MaybeIndex = None, score: MaybeScore = None
    ) -> None:
        """
        Adds a molecule to the chemical space.

        Args:
            mol (MolOrSmiles): The molecule to be added.
                It can be either a Mol object or a SMILES string.
            idx (MaybeIndex, optional): The index of the molecule. Defaults to None.
            score (MaybeScore, optional): The score associated with the molecule.
                Defaults to None.

        Raises:
            ValueError: If scores are enabled and score is not provided.

        Returns:
            None
        """
        mol_m = mol if isinstance(mol, Mol) else utils.smiles2mol(mol)
        self.mols = np.append(self.mols, np.array([mol_m]))

        if self.indices is not None:
            if idx is None:
                raise ValueError("Indices must be provided for all molecules")
            else:
                self.indices = np.append(self.indices, idx)

        if self.scores is not None:
            if score is None:
                raise ValueError("Scores must be provided for all molecules")
            else:
                self.scores = np.append(self.scores, score)

        if self._features is not None:
            feats = np.array(self.featurizer(mol_m)).reshape(1, -1)
            self._features = np.r_[self._features, feats]

    def chunks(self, chunk_size: int) -> Generator[T, None, None]:  # type: ignore
        """
        Split the ChemicalSpaceBaseLayer into chunks of a given size.

        Args:
            chunk_size (int): The size of each chunk.

        Yields:
            Generator[ChemicalSpaceBaseLayer, None, None]:
                A generator of ChemicalSpaceBaseLayer objects.
        """
        for i in range(0, len(self), chunk_size):
            yield self[i : i + chunk_size]

    def deduplicate(self) -> T:  # type: ignore
        """
        Remove duplicate molecules from the ChemicalSpaceBaseLayer by
        keeping only the *first* occurrence of each molecule, based on their InChIKey.

        Returns:
            A new ChemicalSpaceBaseLayer object with duplicate molecules removed.
        """
        cache = set()

        mols_lst: List[Mol] = []
        idx_lst: List[Any] = []
        scores_lst: List[Number] = []
        features_idx: List[int] = []

        mols_hashes = parallel_map(utils.hash_mol, self.mols, n_jobs=self.n_jobs)

        for i in range(len(self)):
            mol = self.mols[i]
            mol_hash = mols_hashes[i]

            if mol_hash in cache:
                continue
            else:
                cache.add(mol_hash)
                mols_lst.append(mol)
                if self.indices is not None:
                    idx_lst.append(self.indices[i])

                if self.scores is not None:
                    scores_lst.append(self.scores[i])

                if self._features is not None:
                    features_idx.append(i)

        idx: Union[List[Any], None] = idx_lst if self.indices is not None else None
        scores: Union[List[Number], None] = (
            scores_lst if self.scores is not None else None
        )

        if self._features is not None:
            features = self._features[features_idx]
        else:
            features = None

        return factory(
            self,
            mols=mols_lst,
            indices=idx,
            scores=scores,
            features=features,
        )

    @cached_property
    def features(self) -> NDArray[np.int_]:
        """
        Calculate the features for each molecule in the chemical space.

        Returns:
            NDArray[int]: An array of features for each molecule.
        """
        if self._features is None:
            self._features = np.array(
                parallel_map(self.featurizer, self.mols, n_jobs=self.n_jobs), dtype=int
            )

        return self._features

    def draw(
        self, indices: bool = True, scores: bool = True, mols_per_row: int = 5, **kwargs
    ):
        """
        Draw the molecules in a grid image.

        Args:
            indices (bool): Whether to include indices in the legends. Defaults to True.
            scores (bool): Whether to include scores in the legends. Defaults to True.
            mols_per_row (int): Number of molecules to display per row. Defaults to 5.
            **kwargs: Additional keyword arguments to be passed to the
                `Draw.MolsToGridImage` function.

        Returns:
            PIL.Image.Image: The grid image containing the drawn molecules.
        """
        from rdkit.Chem import Draw

        Draw.rdDepictor.SetPreferCoordGen(True)

        if indices and (self.indices is not None):
            legends = [str(idx) for idx in self.indices]
        else:
            legends = ["" for _ in range(len(self))]

        if scores and (self.scores is not None):
            legends = [
                f"{legends[i]}\n{float(self.scores[i]):.2f}" for i in range(len(self))
            ]

        # strip initial newline if no indices
        legends = [legend.lstrip("\n") for legend in legends]

        img = Draw.MolsToGridImage(
            self.mols, molsPerRow=mols_per_row, legends=legends, **kwargs
        )
        return img

    @classmethod
    def from_smi(cls: Type[T], path: str, **kwargs) -> T:
        """
        Create a ChemicalSpaceBaseLayer object from a file containing SMILES strings.

        Args:
            path (str): The path to the file containing SMILES strings.
                Can be gzipped
            kwargs (Any): Additional keyword arguments to pass to the constructor.

        Returns:
            ChemicalSpaceBaseLayer: A ChemicalSpaceBaseLayer object
                created from the SMILES strings.

        """

        supplier = utils.smi_supplier(path)

        mols_lst: List[Mol] = []
        indices_lst: List[str] = []

        for mol in supplier:  # type: ignore
            if mol is None:
                warnings.warn("Failed to parse molecule")
                continue

            mols_lst.append(mol)
            indices_lst.append(str(mol.GetProp("_Name")))

        return cls(
            mols=mols_lst,
            indices=indices_lst,
            scores=None,
            **kwargs,
        )

    @classmethod
    def from_sdf(
        cls: Type[T],
        path: str,
        scores_prop: Optional[str] = None,
        cast_to: Callable[[Any], Any] = float,
        **kwargs,
    ) -> T:
        """
        Create a ChemicalSpaceBaseLayer object from an SDF file.

        Args:
            path (str): The path to the SDF file. Can be gzipped.
            scores_prop (Optional[str]): The property name in the SDF file
                that contains the scores. Default is None.
            cast_to (Callable[[Any], Any]): The function to cast the scores to.
                Default is `float`.
            kwargs (Any): Additional keyword arguments to pass to the constructor.

        Returns:
            ChemicalSpaceBaseLayer: The ChemicalSpaceBaseLayer object
                created from the SDF file.

        """
        supplier = utils.sdf_supplier(path)

        mols_lst: List[Mol] = []
        indices_lst: List[str] = []
        scores_lst: List[float] = []

        for mol in supplier:
            if mol is None:
                warnings.warn("Failed to parse molecule")
                continue

            mols_lst.append(mol)
            indices_lst.append(mol.GetProp("_Name"))

            if scores_prop is not None:
                scores_lst.append(cast_to(mol.GetProp(scores_prop)))

        if scores_prop is not None:
            return cls(mols=mols_lst, indices=indices_lst, scores=scores_lst, **kwargs)
        else:
            return cls(
                mols=mols_lst,
                indices=indices_lst,
                scores=None,
                **kwargs,
            )

    def to_smi(self, path: str) -> None:
        """
        Write the molecules in the chemical space to a file as SMILES strings.

        Args:
            path (str): The path to the output file. Can be gzipped

        Returns:
            None
        """
        if self.indices is None:
            indices = [""] * len(self)
        else:
            indices = list(self.indices)

        utils.smi_writer(path, self.mols, indices)

    def to_sdf(self, path: str, scores_prop: Optional[str] = None) -> None:
        """
        Write the molecules in the chemical space to an SDF file.

        Args:
            path (str): The path to the output file. Can be gzipped.
            scores_prop (Optional[str]): The property name to use for the scores.
                Default is None.

        Returns:
            None
        """

        if self.indices is None:
            indices = [""] * len(self)
        else:
            indices = list(map(str, self.indices))

        if scores_prop is not None:
            if self.scores is None:
                raise ValueError("Scores are not available")
            else:
                for mol, idx, score in zip(self.mols, indices, self.scores):
                    mol.SetProp("_Name", idx)
                    mol.SetProp(scores_prop, str(score))
        else:
            for mol, idx in zip(self.mols, indices):
                mol.SetProp("_Name", idx)

        utils.sdf_writer(path, self.mols)

    def __len__(self) -> int:
        """
        Returns the number of molecules in the chemical space.

        Returns:
            int: The number of molecules in the chemical space.
        """
        return len(self.mols)

    def __add__(self, other: T) -> T:
        """
        Adds two ChemicalSpaceBaseLayer objects together.

        Args: other (ChemicalSpaceBaseLayer): The other ChemicalSpaceBaseLayer object
            to add.

        Returns: ChemicalSpaceBaseLayer: A new ChemicalSpaceBaseLayer object that is
            the result of the addition.

        Raises:
            TypeError: If the other object is not an instance of ChemicalSpaceBaseLayer.
        """
        if not isinstance(other, type(self)):
            raise TypeError(f"Can only add {self.name} objects")

        mols = np.concatenate([self.mols, other.mols])
        if (self.indices is None) or (other.indices is None):
            if (self.indices is None) != (other.indices is None):
                warnings.warn(
                    "Both spaces should have indices to concatenate. "
                    "Indices will be None"
                )
            idx = None
        else:
            idx = np.concatenate([self.indices, other.indices])

        if (self.scores is None) or (other.scores is None):
            if (self.scores is None) != (other.scores is None):
                warnings.warn(
                    "One or more spaces do not have scores. Scores will be None"
                )
            score = None
        else:
            score = np.concatenate([self.scores, other.scores])

        if (self._features is None) or (other._features is None):
            # if (self._features is None) != (other._features is None):
            #     warnings.warn(
            #         "One or more spaces do not have features. Features will be None"
            #     )
            features = None
        else:
            features = np.vstack([self._features, other._features])

        return factory(
            self,
            mols=mols,
            indices=idx,
            scores=score,
            features=features,
        )

    def __sub__(self, other: T) -> T:
        """
        Subtract another ChemicalSpaceBaseLayer object from the current object.

        Args:
            other (ChemicalSpaceBaseLayer):
                The ChemicalSpaceBaseLayer object to subtract.

        Returns:
            ChemicalSpaceBaseLayer: A new ChemicalSpaceBaseLayer object
                that contains the molecules from the current object that are not
                present in the other object.

        Raises:
            TypeError: If the other object is not an instance of ChemicalSpaceBaseLayer.
        """
        if not isinstance(other, type(self)):
            raise TypeError("Can only subtract ChemicalSpace objects")

        cache = set(parallel_map(utils.hash_mol, other.mols, n_jobs=self.n_jobs))

        mols_lst: List[Mol] = []
        indices_lst: List[Any] = []
        scores_lst: List[Number] = []
        features_idx: List[int] = []

        mols_hashes = parallel_map(utils.hash_mol, self.mols, n_jobs=self.n_jobs)

        for i in range(len(self)):
            mol: Mol = self.mols[i]
            mol_hash = mols_hashes[i]

            if mol_hash in cache:
                continue
            else:
                mols_lst.append(mol)
                if self.indices is not None:
                    indices_lst.append(self.indices[i])
                if self.scores is not None:
                    scores_lst.append(self.scores[i])
                if self._features is not None:
                    features_idx.append(i)

        idx: Union[List[Any], None] = indices_lst if self.indices is not None else None
        scores: Union[List[Number], None] = (
            scores_lst if self.scores is not None else None
        )
        features = self._features[features_idx] if self._features is not None else None

        return factory(
            self,
            mols=mols_lst,
            indices=idx,
            scores=scores,
            features=features,
            featurizer=self.featurizer,
            metric=self.metric,
            n_jobs=self.n_jobs,
        )

    def __getitem__(
        self, idx: Union[int, SliceType, NDArray[np.bool_], List[bool], List[int]]
    ) -> T:  # type: ignore
        """
        Retrieve the item(s) at the specified index or slice.

        Args:
            idx: The index, slice, mask, or list of indices to retrieve.

        Returns:
            ChemicalSpaceBaseLayer: A new ChemicalSpaceBaseLayer object
                containing the item(s) at the specified index or slice.
        """

        if isinstance(idx, int):
            idx = [idx]  # type: ignore

        mol = self.mols[idx]
        mol_idx = self.indices[idx] if self.indices is not None else None
        score = self.scores[idx] if self.scores is not None else None
        features = self._features[idx] if self._features is not None else None

        return factory(
            self,
            mols=mol,
            indices=mol_idx,
            scores=score,
            features=features,
        )

    def __eq__(self, other: object) -> bool:
        """
        Check if two ChemicalSpaceBaseLayer objects are equal.

        Args:
            other (object): The object to compare with.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        if not isinstance(other, type(self)):
            return False

        if len(self) != len(other):
            return False

        return hash(self) == hash(other)

    def copy(self, deep: bool = False) -> T:  # type: ignore
        """
        Create a copy of the object.

        Args:
            deep (bool): If True, perform a deep copy of the object.
                If False, perform a shallow copy.

        Returns:
            T: A copy of the object.
        """
        if not deep:
            return self.__copy__()
        else:
            return self.__deepcopy__({})

    def __copy__(self) -> T:  # type: ignore
        return factory(
            self,
            mols=self.mols,
            indices=self.indices,
            scores=self.scores,
            features=self._features,
        )

    def __deepcopy__(self, memo: Dict[int, Any]) -> T:  # type: ignore
        _ = memo
        mols = [Mol(mol) for mol in self.mols]
        indices = [idx for idx in self.indices] if self.indices is not None else None
        scores = [score for score in self.scores] if self.scores is not None else None
        features = self._features.copy() if self._features is not None else None
        return factory(
            self,
            mols=mols,
            indices=indices,
            scores=scores,
            features=features,
        )

    @staticmethod
    def hash_mol(mol: Mol) -> str:
        """
        Compute the hash of a molecule.

        Args:
            mol (Mol): The molecule to hash.

        Returns:
            str: The hash of the molecule.
        """
        return Chem.MolToInchiKey(mol)

    def __hash__(self) -> int:
        inchi_keys: List[str] = parallel_map(
            self.hash_mol, self.mols, n_jobs=self.n_jobs
        )

        if self.hash_indices:
            indices = self.indices if self.indices is not None else [""] * len(self)
            mol_strings = [
                f"{inchi_key}@{idx}" for inchi_key, idx in zip(inchi_keys, indices)
            ]
        else:
            mol_strings = inchi_keys

        return hash(frozenset(mol_strings))

    def __repr__(self) -> str:
        idx_repr = len(self.indices) if self.indices is not None else "No"
        scores_repr = len(self.scores) if self.scores is not None else "No"
        return (
            f"<{self.name}: {len(self)} molecules "
            f"| {idx_repr} indices | {scores_repr} scores>"
        )

    def __str__(self) -> str:
        return self.__repr__()



from functools import reduce
from typing import (
    Any,
    Callable,
    Generator,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
from joblib import Parallel, delayed
from numpy.typing import NDArray
from rdkit import Chem
from rdkit.Chem import AllChem, Mol, inchi  # type: ignore
from typing_extensions import TypeAlias

Number: TypeAlias = Union[int, float]
MaybeIndex: TypeAlias = Union[Any, None]
MaybeScore: TypeAlias = Union[Number, None]
MolOrSmiles: TypeAlias = Union[Mol, str]
IntOrNone: TypeAlias = Union[int, None]
SliceType: TypeAlias = slice
MolFeaturizerType: TypeAlias = Callable[
    [Mol], Union[Sequence[Number], NDArray[np.int_], NDArray[np.float_]]
]
ArrayIntOrFloat: TypeAlias = Union[NDArray[np.int_], NDArray[np.float_]]


SEED: int = 42


def reduce_sum(*objects: Any) -> Any:
    """
    Returns the sum of all the objects passed as arguments.

    Args:
        *objects: Variable number of objects to be summed.

    Returns:
        The sum of all the objects.

    """
    return reduce(lambda x, y: x + y, objects)


def hash_mol(mol: Mol) -> str:
    """
    Returns the InChIKey hash of a given molecule.

    Args:
        mol (Mol): The molecule to be hashed.

    Returns:
        str: The InChIKey hash of the molecule.
    """
    return inchi.MolToInchiKey(mol)


def ecfp4_featurizer(mol: Mol, radius: int = 2, n_bits: int = 1024) -> List[int]:
    """
    Calculates the ECFP4 fingerprint for a given molecule.

    Args:
        mol (rdkit.Chem.Mol): The molecule to calculate the fingerprint for.
        radius (int, optional): The radius of the fingerprint. Defaults to 2.
        n_bits (int, optional): The number of bits in the fingerprint. Defaults to 1024.

    Returns:
        List[int]: The ECFP4 fingerprint as a list of integers.
    """
    return AllChem.GetMorganFingerprintAsBitVect(  # type: ignore
        mol, radius, nBits=n_bits, useChirality=True
    ).ToList()


def maccs_featurizer(mol: Mol) -> List[int]:
    """
    Calculates the MACCS fingerprint for a given molecule.

    Args:
        mol (rdkit.Chem.Mol): The molecule to calculate the fingerprint for.

    Returns:
        List[int]: The MACCS fingerprint as a list of integers.
    """
    return list(AllChem.GetMACCSKeysFingerprint(mol))  # type: ignore


def smiles2mol(smiles: str) -> Mol:
    """
    Convert a SMILES string to a molecule object.

    Args:
        smiles (str): The SMILES string representing the molecule.

    Returns:
        Mol: The molecule object generated from the SMILES string.

    Raises:
        ValueError: If the SMILES string cannot be parsed into a molecule object.
    """
    mol = Chem.MolFromSmiles(smiles)  # type: ignore
    if mol is None:
        raise ValueError(f"Failed to parse SMILES: {smiles}")
    return mol


def safe_smiles2mol(smiles_or_mol: MolOrSmiles) -> Mol:
    """
    Convert a SMILES string or RDKit Mol object to an RDKit Mol object.

    Args:
        smiles_or_mol (MolOrSmiles): A RDKit Mol object or a SMILES string.

    Returns:
        Mol: The RDKit Mol object.

    """
    if isinstance(smiles_or_mol, Mol):
        return smiles_or_mol
    return smiles2mol(smiles_or_mol)


def factory(cls, **kwargs) -> Any:
    """
    Create a new instance of the given class with the provided arguments.
    Any missing arguments to the __init__ in kwargs will be filled with the
    class attributes.

    Args:
        cls: The class object to create a new instance of.
        **kwargs: Keyword arguments to pass to the class constructor.

    Returns:
        An instance of the given class.

    """

    init_arg_names = get_class_init_args(cls)

    # Isolate init arguments that were not passed
    missing_args = set(init_arg_names) - set(kwargs.keys())
    # And pull them from the instance
    missing_kwargs = {arg: getattr(cls, arg) for arg in missing_args}

    # Combine the provided and missing arguments
    kwargs = {**kwargs, **missing_kwargs}

    obj = type(cls)(**kwargs)

    return obj


def get_class_init_args(obj: Any) -> Tuple[str]:
    """
    Get the argument names of the class __init__ method.

    Args:
        obj: The class instance.

    Returns:
        Tuple[str]: The argument names of the class __init__ method.

    """

    return obj.__init__.__code__.co_varnames[1 : obj.__init__.__code__.co_argcount]


def parallel_map(
    func: Callable, iterable: Iterable[Any], n_jobs: int = -1
) -> List[Any]:
    """
    Apply a function to each item in the iterable in parallel using multiple processes.

    Args:
        func (Callable): The function to apply to each item.
        iterable (Iterable[Any]): The iterable containing the
            items to apply the function to.
        n_jobs (int, optional): The number of processes to use for parallel execution.
            Defaults to -1, which uses all available processors.

    Returns:
        List[Any]: A list containing the results of applying the function
            to each item in the iterable.
    """
    return list(Parallel(n_jobs=n_jobs)(delayed(func)(item) for item in iterable))


def smi_supplier(path: str) -> Generator[Mol, None, None]:
    """
    Generator function that reads a file containing SMILES strings and
    yields RDKit molecules.

    Args:
        path (str): The path to the file containing SMILES strings.
            Can be a .smi, .smiles, or .smi.gz file.

    Yields:
        Chem.Mol: RDKit molecule object.

    """
    if path.lower().endswith("gz"):
        import gzip

        f = gzip.open(path, "rt")
    else:
        f = open(path, "r")

    for line in f:
        fields = line.strip().split()
        if len(fields) > 1:
            name = fields[1]
        else:
            name = ""

        mol = Chem.MolFromSmiles(fields[0])  # type: ignore

        if mol is not None:
            mol.SetProp("_Name", name)
            yield mol

    f.close()


def smi_writer(
    path: str, mols: Iterable[Mol], names: Optional[Sequence[Any]] = None
) -> None:
    """
    Write a list of RDKit molecules to a file in SMILES format.

    Args:
        path (str): The path to the output file. Can be a .smi, .smiles, or .smi.gz file
        mols (Iterable[Mol]): The list of RDKit molecules to write to the file
        names (Optional[Sequence[Any]], optional): The names of the molecules
            If None, the names are taken from the RDKit molecule properties.
            Defaults to None
    """

    if path.lower().endswith("gz"):
        import gzip

        handle = gzip.open(path, "wt")
    else:
        handle = open(path, "w")

    for i, mol in enumerate(mols):
        smiles = Chem.MolToSmiles(mol)  # type: ignore
        if names is not None:
            name = str(names[i])
        else:
            name = mol.GetProp("_Name")
        handle.write(f"{smiles} {name}\n")

    handle.close()


def sdf_supplier(path: str, **kwargs) -> Generator[Mol, None, None]:
    """
    Generator function that reads an SDF file and yields RDKit molecules.

    Args:
        path (str): The path to the SDF file. Can be a .sdf or .sdf.gz file.
        **kwargs: Additional keyword arguments to be passed
            to the RDKit ForwardSDMolSupplier.

    Yields:
        Chem.Mol: RDKit molecule object.

    Returns:
        None
    """
    if path.lower().endswith("gz"):
        import gzip

        handle = gzip.open(path, "rb")
    else:
        handle = open(path, "rb")  # type: ignore

    supplier = Chem.ForwardSDMolSupplier(handle, **kwargs)  # type: ignore

    for mol in supplier:
        if mol is not None:
            yield mol

    handle.close()


def sdf_writer(path: str, mols: Iterable[Mol], **kwargs) -> None:
    """
    Write a list of RDKit molecules to a file in SDF format.

    Args:
        path (str): The path to the output file. Can be a .sdf or .sdf.gz file.
        mols (Iterable[Mol]): The list of RDKit molecules to write to the file.
        **kwargs: Additional keyword arguments to be passed to the RDKit SDWriter.
    """

    if path.lower().endswith("gz"):
        import gzip

        f = gzip.open(path, "wt")
    else:
        f = open(path, "wt")

    writer = Chem.SDWriter(f, **kwargs)  # type: ignore
    for mol in mols:
        writer.write(mol)

    # Somehow necessary to flush the writer for pytest to work
    writer.flush()  # type: ignore
    f.flush()

    f.close()



import os
from typing import Generator, Optional, Tuple

import numpy as np
import pytest

from chemicalspace.layers.clustering import (
    CLUSTERING_METHODS,
    CLUSTERING_METHODS_N,
    ChemicalSpaceClusteringLayer,
    ClusteringMethodsType,
    ClusteringMethodsTypeN,
    get_optimal_cluster_number,
)

np.random.seed(42)

TESTS_DIR = os.path.dirname(os.path.abspath(__file__))

INPUT_SMI_FILE = os.path.join(TESTS_DIR, "data", "inputs1.smi")


@pytest.fixture
def space() -> ChemicalSpaceClusteringLayer:
    return ChemicalSpaceClusteringLayer.from_smi(INPUT_SMI_FILE)


@pytest.mark.parametrize("method", CLUSTERING_METHODS_N)
def test_silhouette(
    space: ChemicalSpaceClusteringLayer, method: ClusteringMethodsTypeN
) -> None:
    from functools import partial

    from sklearn.cluster import AgglomerativeClustering
    from sklearn_extra.cluster import KMedoids

    if method == "kmedoids":
        obj = partial(KMedoids, metric="jaccard", random_state=42)
    elif method == "agglomerative-clustering":
        obj = partial(AgglomerativeClustering, metric="jaccard", linkage="complete")
    else:
        raise ValueError(f"Invalid clustering method: {method}")

    n_clusters = get_optimal_cluster_number(space.features, model=obj)

    if method == "kmedoids":
        assert n_clusters == 3
    elif method == "agglomerative-clustering":
        assert n_clusters == 5
    else:
        raise ValueError(f"Invalid clustering method: {method}")


@pytest.mark.parametrize("method", CLUSTERING_METHODS_N)
def test_cluster_n(
    space: ChemicalSpaceClusteringLayer,
    method: ClusteringMethodsTypeN,
    n_clusters: Optional[int] = 3,
) -> None:

    clusters = space.cluster(n_clusters=n_clusters, method=method, seed=42)

    assert isinstance(clusters, np.ndarray)
    assert np.issubdtype(clusters.dtype, np.integer)
    assert len(clusters) == len(space)

    if n_clusters is not None:
        assert set(clusters) == set(range(n_clusters))


@pytest.mark.parametrize("method", CLUSTERING_METHODS)
def test_cluster(
    space: ChemicalSpaceClusteringLayer, method: ClusteringMethodsType
) -> None:

    clusters = space.cluster(method=method, seed=42)

    assert isinstance(clusters, np.ndarray)
    assert np.issubdtype(clusters.dtype, np.integer)
    assert len(clusters) == len(space)


@pytest.mark.parametrize(
    "method, n_clusters", [("kmedoids", 3), ("agglomerative-clustering", None)]
)
def test_yield_clusters_n(
    space: ChemicalSpaceClusteringLayer,
    method: ClusteringMethodsTypeN,
    n_clusters: Optional[int],
) -> None:
    clusters: Generator[ChemicalSpaceClusteringLayer, None, None] = (
        space.yield_clusters(n_clusters=n_clusters, method=method, seed=42)
    )

    i = None
    for i, cluster in enumerate(clusters):
        assert isinstance(cluster, ChemicalSpaceClusteringLayer)
        assert len(cluster) > 0

    if n_clusters is not None:
        assert i == n_clusters - 1


@pytest.mark.parametrize(
    "method, kwargs", [("sphere-exclusion", {"radius": 0.8}), ("scaffold", {})]
)
def test_yield_clusters(
    space: ChemicalSpaceClusteringLayer, method: ClusteringMethodsType, kwargs
) -> None:
    clusters: Generator[ChemicalSpaceClusteringLayer, None, None] = (
        space.yield_clusters(method=method, seed=42, **kwargs)
    )

    i = None
    for i, cluster in enumerate(clusters):
        assert isinstance(cluster, ChemicalSpaceClusteringLayer)
        assert len(cluster) > 0

    assert i is not None


@pytest.mark.parametrize(
    "method, n_clusters", [("kmedoids", 3), ("agglomerative-clustering", 3)]
)
def test_split(
    space: ChemicalSpaceClusteringLayer,
    method: ClusteringMethodsTypeN,
    n_clusters: int,
) -> None:
    ks: Generator[
        Tuple[ChemicalSpaceClusteringLayer, ChemicalSpaceClusteringLayer], None, None
    ] = space.split(n_splits=n_clusters, method=method, seed=42)

    for cluster_train, cluster_test in ks:

        assert isinstance(cluster_train, ChemicalSpaceClusteringLayer)
        assert isinstance(cluster_test, ChemicalSpaceClusteringLayer)

        assert len(cluster_train) > 0
        assert len(cluster_test) > 0
        # Check that there's no overlap between train and test clusters
        assert cluster_train - cluster_test == cluster_train
        assert cluster_train + cluster_test == space



import os

import numpy as np
import pytest

from chemicalspace.layers.projection import (
    PROJECTION_METHODS,
    ChemicalSpaceProjectionLayer,
)

np.random.seed(42)

TESTS_DIR = os.path.dirname(os.path.abspath(__file__))

INPUT_SMI_FILE = os.path.join(TESTS_DIR, "data", "inputs1.smi")


@pytest.fixture
def space() -> ChemicalSpaceProjectionLayer:
    return ChemicalSpaceProjectionLayer.from_smi(INPUT_SMI_FILE)


@pytest.mark.parametrize("n_components", (2,))
@pytest.mark.parametrize("method", ("pca", "tsne"))  # umap is slow
def test_project(
    space: ChemicalSpaceProjectionLayer, n_components: int, method: PROJECTION_METHODS
) -> None:
    if method == "tsne":
        kwargs = {"perplexity": 3}
    elif method == "umap":
        kwargs = {"n_neighbors": 3}
    else:
        kwargs = {}

    proj = space.project(n_components=n_components, method=method, n_jobs=1, **kwargs)
    assert isinstance(proj, np.ndarray)
    assert proj.shape == (10, n_components)



from .layers.acquisition import ChemicalSpaceAcquisitionLayer
from .layers.clustering import ChemicalSpaceClusteringLayer
from .layers.diversity import ChemicalSpaceDiversityLayer
from .layers.neighbors import ChemicalSpaceNeighborsLayer
from .layers.projection import ChemicalSpaceProjectionLayer


class ChemicalSpace(
    ChemicalSpaceClusteringLayer,
    ChemicalSpaceNeighborsLayer,
    ChemicalSpaceProjectionLayer,
    ChemicalSpaceAcquisitionLayer,
    ChemicalSpaceDiversityLayer,
):
    """
    ChemicalSpace is a class that combines different layers to create a
    complete representation of a chemical space"""

    pass



from abc import ABC, abstractmethod
from typing import Any, List, Literal, Optional, Sequence, Union

import numpy as np
from numpy.typing import NDArray
from typing_extensions import TypeAlias

from chemicalspace.layers.base import ChemicalSpaceBaseLayer, T


class BaseAcquisitionStrategy(ABC):
    """
    BaseAcquisitionStrategy is an abstract class that defines the interface
    for acquisition strategies. Acquisition strategies are classes that
    implement a method to pick samples from a set of inputs.
    """

    def __init__(self, inputs: Union[Sequence[Any], NDArray[Any]], **kwargs):
        """
        Initialize the acquisition strategy with input object features and scores

        Args:
            inputs (Sequence[Any] | NDArray[Any]): Inputs features to the
                acquisition strategy.
            scores (Optional[Sequence[Any] | NDArray[Any]], optional): Scores
                associated with the inputs. Defaults to None.
        """
        self.inputs = np.array(inputs)
        self.kwargs = kwargs

    @abstractmethod
    def __call__(self, n: int) -> List[int]:
        """
        Pick samples from the inputs.

        Args:
            n (int): Number of samples to pick.

        Returns:
            List[int]: List of indices of the picked samples.

        """
        raise NotImplementedError


class BaseAcquisitionStrategyRequiresScores(BaseAcquisitionStrategy, ABC):
    """
    BaseAcquisitionStrategyRequiresScores is an abstract class that defines the
    interface for acquisition strategies that require scores to pick samples.
    """

    def __init__(
        self,
        inputs: Union[Sequence[Any], NDArray[Any]],
        scores: Union[Sequence[Any], NDArray[Any]],
        **kwargs,
    ):
        """
        Initialize the acquisition strategy with input object features and scores

        Args:
            inputs (Sequence[Any] | NDArray[Any]):
                Inputs features to the acquisition strategy.
            scores (Optional[Sequence[Any] | NDArray[Any]], optional):
                Scores associated with the inputs. Defaults to None.
        """
        super().__init__(inputs, **kwargs)
        self.scores = scores


class RandomStrategy(BaseAcquisitionStrategy):
    """
    RandomStrategy is an acquisition strategy that picks samples randomly.
    """

    def __call__(self, n: int) -> List[int]:
        """
        Pick n samples from the inputs randomly.

        Args:
            n (int): Number of samples to pick.

        Returns:
            List[int]: List of indices of the picked samples.

        """
        idx = np.random.choice(len(self.inputs), n, replace=False)
        return idx.tolist()


class GreedyStrategy(BaseAcquisitionStrategyRequiresScores):
    """
    GreedyStrategy is an acquisition strategy
    that picks samples greedily based on scores.
    """

    def __call__(self, n: int) -> List[int]:
        """
        Pick the top n samples from the inputs based on scores.

        Args:
            n (int): Number of samples to pick.

        Returns:
            List[int]: List of indices of the picked samples.

        """
        idx = np.argsort(self.scores)[-n:]
        return idx.tolist()


class MaxMinStrategy(BaseAcquisitionStrategy):
    """
    MaxMinStrategy is an acquisition strategy that picks
    samples based on the MaxMin algorithm.
    """

    def __init__(
        self,
        inputs: Union[Sequence[Any], NDArray[Any]],
        metric: str = "jaccard",
        seed: int = 42,
        **kwargs,
    ):
        """
        Initialize the acquisition strategy with input object features and scores

        Args:
            inputs (Sequence[Any] | NDArray[Any]): Inputs features to the
                acquisition strategy.
            metric (str): Metric to use for computing the distance between the samples.
            seed (int, optional): Random seed for reproducibility. Defaults to 42.
        """
        super().__init__(inputs, **kwargs)
        self.metric = metric

        from scipy.spatial import distance

        try:
            self.metric_fn = getattr(distance, metric)
        except AttributeError:
            raise ValueError(
                f"Unknown metric: {metric}. "
                f"Pass a valid metric from scipy.spatial.distance."
            )

        def distij(i: int, j: int) -> float:
            return self.metric_fn(self.inputs[i], self.inputs[j])

        self.distij = distij

        self.seed = seed

    def __call__(self, n: int) -> List[int]:
        """
        Pick the top n samples from the inputs based on the maximum minimum distance.

        Args:
            n (int): Number of samples to pick.

        Returns:
            List[int]: List of indices of the picked samples.

        """
        from rdkit.SimDivFilters.rdSimDivPickers import MaxMinPicker

        picker = MaxMinPicker()  # type: ignore
        indices = picker.LazyPick(self.distij, len(self.inputs), n, seed=self.seed)

        return list(indices)


STRATEGIES: TypeAlias = Literal["random", "greedy", "maxmin"]
strategies_dict = {
    "random": RandomStrategy,
    "greedy": GreedyStrategy,
    "maxmin": MaxMinStrategy,
}


def pick_samples(
    n: int,
    strategy: STRATEGIES,
    inputs: Union[Sequence[Any], NDArray[Any]],
    scores: Optional[Union[Sequence[Any], NDArray[Any]]] = None,
    **strategy_kwargs,
) -> NDArray[np.int_]:
    """
    Pick n samples from the inputs using the specified strategy.

    Args:
        n (int): Number of samples to pick.
        strategy (STRATEGIES): Strategy to use for picking samples.
        inputs (Sequence[Any] | NDArray[Any]):
            Inputs features to the acquisition strategy.
        scores (Optional[Sequence[Any] | NDArray[Any]], optional):
            Scores associated with the inputs.
        **strategy_kwargs: Additional keyword arguments
            to pass to the acquisition strategy.

    Returns:
        NDArray[int]: Array of indices of the picked samples.

    """
    if strategy not in strategies_dict:
        raise ValueError(
            f"Unknown strategy: {strategy}. "
            f"Allowed strategies: {list(strategies_dict.keys())}"
        )

    strategy_class = strategies_dict[strategy]
    strategy_instance = strategy_class(inputs=inputs, scores=scores, **strategy_kwargs)

    return np.array(strategy_instance(n), dtype=int)


class ChemicalSpaceAcquisitionLayer(ChemicalSpaceBaseLayer):
    """
    ChemicalSpaceAcquisitionLayer is a class that provides acquisition
    functionalities to a chemical space. It is a mixin class that can be
    combined with other layers to create a complete representation of a
    chemical space.
    """

    def pick(
        self, n: int, strategy: STRATEGIES = "random", **strategy_kwargs
    ) -> T:  # type: ignore
        """
        Pick n samples from the chemical space using the specified strategy.

        Args:
            n (int): Number of samples to pick.
            strategy (STRATEGIES, optional): Strategy to use for picking samples.
                Defaults to "random".
            **strategy_kwargs: Additional keyword arguments to pass
                to the acquisition strategy.

        Returns:
            T: A ChemicalSpace object with the picked samples.

        """
        pick_idx = pick_samples(
            n=n,
            strategy=strategy,
            inputs=self.features,
            scores=self.scores,
            **strategy_kwargs,
        )
        mask = np.zeros(len(self.features), dtype=bool)
        mask[pick_idx] = True

        return self[mask]



import os
from pathlib import Path
from types import GeneratorType
from typing import Any, Generator, List, Optional

import numpy as np
import pytest
from rdkit import Chem
from rdkit.Chem.rdchem import Mol

from chemicalspace.layers.base import ChemicalSpaceBaseLayer
from chemicalspace.utils import (
    MolFeaturizerType,
    Number,
    ecfp4_featurizer,
    maccs_featurizer,
)

TESTS_DIR = os.path.dirname(os.path.abspath(__file__))

INPUT_SMI_FILES = [
    os.path.join(TESTS_DIR, "data", name) for name in ["inputs1.smi", "inputs2.smi.gz"]
]
INPUT_SDF_FILES = [
    os.path.join(TESTS_DIR, "data", name) for name in ["inputs1.sdf", "inputs2.sdf.gz"]
]

test_smiles = ["C", "CC", "CCC", "CCCC", "CCCCC", "CCCCCC", "CCCCCCC"]
test_mols = [Chem.MolFromSmiles(smi) for smi in test_smiles]  # type: ignore


@pytest.fixture
def space() -> ChemicalSpaceBaseLayer:
    return ChemicalSpaceBaseLayer.from_smi(INPUT_SMI_FILES[0])


@pytest.fixture
def space_with_scores() -> ChemicalSpaceBaseLayer:
    scores = np.array(range(len(INPUT_SMI_FILES[0])))
    return ChemicalSpaceBaseLayer.from_smi(INPUT_SMI_FILES[0], scores=scores)


@pytest.fixture
def other_space() -> ChemicalSpaceBaseLayer:
    return ChemicalSpaceBaseLayer.from_smi(INPUT_SMI_FILES[1])


@pytest.mark.parametrize(
    "scores", [None, [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7], [1, 2, 3, 4, 5, 6, 7]]
)
@pytest.mark.parametrize(
    "indices", [None, ["a", "b", "c", "d", "e", "f", "g"], [1, 2, 3, 4, 5, 6, 7]]
)
@pytest.mark.parametrize("smiles_or_mols_lst", [[], test_smiles, test_mols])
def test_initialization(
    smiles_or_mols_lst: List[str],
    indices: Optional[List[Any]],
    scores: Optional[List[Number]],
) -> None:

    should_raise: bool = (
        (indices is not None) and (len(indices) != len(smiles_or_mols_lst))
    ) or ((scores is not None) and (len(scores) != len(smiles_or_mols_lst)))
    if should_raise:
        with pytest.raises(ValueError):
            space = ChemicalSpaceBaseLayer(
                smiles_or_mols_lst, indices=indices, scores=scores
            )
        return
    else:
        space = ChemicalSpaceBaseLayer(
            smiles_or_mols_lst, indices=indices, scores=scores
        )

    assert len(space) == len(smiles_or_mols_lst)
    if indices is not None:
        assert len(space) == len(indices)
    if scores is not None:
        assert len(space) == len(scores)


@pytest.mark.parametrize(
    "input_file,featurizer,metric",
    [
        (INPUT_SMI_FILES[0], ecfp4_featurizer, "jaccard"),
        (INPUT_SMI_FILES[0], maccs_featurizer, "jaccard"),
        (INPUT_SDF_FILES[0], ecfp4_featurizer, "euclidean"),
        (INPUT_SDF_FILES[0], maccs_featurizer, "euclidean"),
    ],
)
def test_classmethods(
    input_file: str, featurizer: MolFeaturizerType, metric: str
) -> None:
    if input_file.endswith(".smi"):
        space = ChemicalSpaceBaseLayer.from_smi(
            input_file, featurizer=featurizer, metric=metric
        )
    else:
        space = ChemicalSpaceBaseLayer.from_sdf(
            input_file, featurizer=featurizer, metric=metric
        )

    assert len(space) == 10

    assert len(space.mols) == 10
    assert isinstance(space.mols, np.ndarray)
    assert isinstance(space.mols[0], Mol)

    assert space.indices is not None
    assert len(space.indices) == 10
    assert isinstance(space.indices[0], str)

    assert space.scores is None


@pytest.mark.parametrize("gzipped", [False, True])
def test_smi_io(space: ChemicalSpaceBaseLayer, tmp_path: Path, gzipped: bool) -> None:
    if gzipped:
        fname = "test.smi.gz"
    else:
        fname = "test.smi"

    space.to_smi(str(tmp_path / fname))

    assert (tmp_path / fname).exists()

    space_loaded = ChemicalSpaceBaseLayer.from_smi(str(tmp_path / fname))

    assert space == space_loaded


@pytest.mark.parametrize("scores_prop", [None, "scores"])
@pytest.mark.parametrize("gzipped", [False, True])
@pytest.mark.parametrize("add_scores", [True, False])
def test_sdf_io(
    space: ChemicalSpaceBaseLayer,
    tmp_path: Path,
    gzipped: bool,
    add_scores: bool,
    scores_prop: Optional[str],
) -> None:

    if gzipped:
        fname = "test.sdf.gz"
    else:
        fname = "test.sdf"

    if add_scores:
        space.scores = np.array(range(len(space)))

    if (scores_prop is not None) and (space.scores is None):
        with pytest.raises(ValueError):
            space.to_sdf(str(tmp_path / fname), scores_prop=scores_prop)
        return
    else:
        space.to_sdf(str(tmp_path / fname), scores_prop=scores_prop)

    assert (tmp_path / fname).exists()

    space_loaded = ChemicalSpaceBaseLayer.from_sdf(
        str(tmp_path / fname), scores_prop=scores_prop
    )

    assert space == space_loaded

    if (space.indices is not None) and (space_loaded.indices is not None):
        assert (space.indices == space_loaded.indices).all()

    if scores_prop is not None:
        if (space.scores is not None) and (space_loaded.scores is not None):
            assert (space.scores == space_loaded.scores).all()
        else:
            raise ValueError("Scores None for one of the spaces")


@pytest.mark.parametrize("featurizer", [ecfp4_featurizer, maccs_featurizer])
def test_features(space: ChemicalSpaceBaseLayer, featurizer: MolFeaturizerType) -> None:
    space.featurizer = featurizer

    assert space._features is None
    assert isinstance(space.features, np.ndarray)

    if featurizer == ecfp4_featurizer:
        assert space.features.shape == (10, 1024)
    elif featurizer == maccs_featurizer:
        assert space.features.shape == (10, 167)
    else:
        raise ValueError("Invalid featurizer")

    # Check that the features are cached
    assert space._features is not None


def test_slice_mask_chunk(space: ChemicalSpaceBaseLayer) -> None:
    # Compute the features
    _ = space.features

    entry: ChemicalSpaceBaseLayer = space[0]
    assert isinstance(entry, type(space))
    assert isinstance(entry.mols[0], Mol)
    if entry.indices is not None:
        assert isinstance(entry.indices[0], str)
    assert entry.scores is None

    entries: ChemicalSpaceBaseLayer = space[:3]

    assert isinstance(entries, type(space))
    assert len(entries) == 3
    assert entries.scores is None
    assert isinstance(entries.mols, np.ndarray)
    assert isinstance(entries.mols[0], Mol)
    assert isinstance(entries.indices[0], str)  # type: ignore

    space_slice: ChemicalSpaceBaseLayer = space[0:10:2]

    assert isinstance(space_slice, ChemicalSpaceBaseLayer)
    assert len(space_slice) == 5
    # Check features, but bypass `.features` to check if the array has
    # been assigned when instantiating
    assert np.allclose(space_slice._features, space._features[::2])  # type: ignore

    mask = [True, False] * 5
    space_mask: ChemicalSpaceBaseLayer = space[mask]

    assert isinstance(space_mask, ChemicalSpaceBaseLayer)
    assert len(space_mask) == 5
    assert np.allclose(space_mask._features, space._features[::2])  # type: ignore

    assert space_slice == space_mask

    space_chunks: Generator[ChemicalSpaceBaseLayer, None, None] = space.chunks(
        chunk_size=3
    )
    space_chunks_lst = list(space_chunks)
    space_chunks_lst_sizes = [len(chunk) for chunk in space_chunks_lst]

    assert isinstance(space_chunks, GeneratorType)
    assert len(space_chunks_lst) == 4
    assert space_chunks_lst_sizes == [3, 3, 3, 1]
    assert space_chunks_lst[0] == space[0:3]
    assert np.allclose(
        space_chunks_lst[0]._features, space._features[:3]  # type: ignore
    )
    assert np.allclose(
        space_chunks_lst[-1]._features, space._features[-1:]  # type: ignore
    )


def test_attribute_inheritance(space: ChemicalSpaceBaseLayer) -> None:
    """
    Test that the class factory properly propagates attributes to the new class.
    """

    class DerivedSpace(ChemicalSpaceBaseLayer):
        def __init__(self, new_parameter: int, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.new_parameter = new_parameter

    derived_space = DerivedSpace(
        new_parameter=42, mols=space.mols, indices=space.indices
    )

    derived_space_slice: DerivedSpace = derived_space[:5]
    assert derived_space_slice.new_parameter == 42

    derived_space_mask: DerivedSpace = derived_space[[True, False] * 5]
    assert derived_space_mask.new_parameter == 42

    derived_space_chunks: Generator[DerivedSpace, None, None] = derived_space.chunks(3)
    for chunk in derived_space_chunks:
        assert chunk.new_parameter == 42

    derived_space_copy: DerivedSpace = derived_space.copy()
    assert derived_space_copy.new_parameter == 42

    derived_space_add: DerivedSpace = derived_space + derived_space
    assert derived_space_add.new_parameter == 42

    derived_space_sub: DerivedSpace = derived_space - derived_space
    assert derived_space_sub.new_parameter == 42

    derived_space_dedup: DerivedSpace = derived_space.deduplicate()
    assert derived_space_dedup.new_parameter == 42


def test_copy(space: ChemicalSpaceBaseLayer) -> None:
    space_shallow: ChemicalSpaceBaseLayer = space.copy(deep=False)
    assert space == space_shallow
    assert id(space.mols[0]) == id(space_shallow.mols[0])

    space_deep: ChemicalSpaceBaseLayer = space.copy(deep=True)
    assert space == space_deep
    assert id(space.mols[0]) != id(space_deep.mols[0])


@pytest.mark.parametrize("use_indices", [True, False])
def test_hashing(use_indices: bool, input_file: str = INPUT_SMI_FILES[0]) -> None:
    space = ChemicalSpaceBaseLayer.from_smi(input_file, hash_indices=use_indices)
    space_noindices = ChemicalSpaceBaseLayer.from_smi(input_file, hash_indices=False)

    assert space == space
    if use_indices:
        assert space != space_noindices
    else:
        assert space == space_noindices


def test_deduplicate(space: ChemicalSpaceBaseLayer) -> None:
    space_twice = space + space

    space_dedup: ChemicalSpaceBaseLayer = space_twice.deduplicate()
    assert len(space_dedup) == len(space)
    assert space_dedup == space


def test_dual_operations(
    space: ChemicalSpaceBaseLayer, other_space: ChemicalSpaceBaseLayer
) -> None:
    # Compute the features
    _ = space.features, other_space.features

    assert space != other_space
    assert space == space
    assert other_space == other_space

    if other_space.indices is None:
        idx = None
    else:
        idx = other_space.indices[0]

    space_add: ChemicalSpaceBaseLayer = space.copy()
    space_add.add(mol=other_space.mols[0], idx=idx)
    assert len(space_add) == len(space) + 1
    # Placate pyright
    if (
        (space._features is not None)
        and (other_space._features is not None)
        and (space_add._features is not None)
    ):
        assert np.allclose(
            space_add._features,
            np.vstack([space._features, other_space._features[0]]),  # type: ignore
        )
    else:
        raise ValueError("Features are not computed")

    combined_spaces = space + other_space

    assert len(combined_spaces) == 10 + 15
    assert combined_spaces[0:10] == space
    assert combined_spaces[10:] == other_space
    assert combined_spaces._features is not None
    # Placate pyright
    if (combined_spaces._features is not None) and (other_space._features is not None):
        assert np.allclose(
            combined_spaces._features,
            np.vstack([space._features, other_space._features]),  # type: ignore
        )
    else:
        raise ValueError("Features are not computed")

    subtracted_spaces = other_space - space
    assert len(subtracted_spaces) == 10
    assert subtracted_spaces._features is not None
    assert np.allclose(
        subtracted_spaces._features, other_space._features[:-5]  # type: ignore
    )

    empty_space = space - space
    assert len(empty_space) == 0
    assert empty_space._features is not None
    assert empty_space._features.size == 0


@pytest.mark.parametrize("scores", [True, False])
@pytest.mark.parametrize("indices", [True, False])
def test_draw(
    space: ChemicalSpaceBaseLayer, tmp_path: Path, indices: bool, scores: bool
) -> None:
    img = space.draw(indices=indices, scores=scores)
    img.save(str(tmp_path / "test.png"))

    assert (tmp_path / "test.png").exists()


def test_repr(space: ChemicalSpaceBaseLayer) -> None:
    assert (
        repr(space) == "<ChemicalSpaceBaseLayer: 10 molecules | 10 indices | No scores>"
    )


def test_str(space: ChemicalSpaceBaseLayer) -> None:
    assert (
        str(space) == "<ChemicalSpaceBaseLayer: 10 molecules | 10 indices | No scores>"
    )



import os

import numpy as np
import pytest

from chemicalspace.layers.neighbors import ChemicalSpaceNeighborsLayer

np.random.seed(42)

TESTS_DIR = os.path.dirname(os.path.abspath(__file__))

INPUT_SMI_FILES = [
    os.path.join(TESTS_DIR, "data", name) for name in ["inputs1.smi", "inputs2.smi.gz"]
]


@pytest.fixture
def space() -> ChemicalSpaceNeighborsLayer:
    return ChemicalSpaceNeighborsLayer.from_smi(INPUT_SMI_FILES[0])


@pytest.fixture
def other_space() -> ChemicalSpaceNeighborsLayer:
    return ChemicalSpaceNeighborsLayer.from_smi(INPUT_SMI_FILES[1])


def test_overlap(space, other_space, radius=0.4, min_neighbors=1):
    indices = space.find_overlap(
        other_space, radius=radius, min_neighbors=min_neighbors
    )
    assert isinstance(indices, np.ndarray)
    assert len(indices) == 5


def test_carve(space, other_space, radius=0.4, min_neighbors=1):
    other_space_carved = other_space.carve(
        space, radius=radius, min_neighbors=min_neighbors
    )
    assert len(other_space - other_space_carved) == 5



from functools import lru_cache
from typing import Literal

import numpy as np
from numpy.typing import NDArray
from typing_extensions import TypeAlias

from chemicalspace.utils import SEED

from .base import ChemicalSpaceBaseLayer

PROJECTION_METHODS: TypeAlias = Literal["umap", "tsne", "pca"]


def project_space(
    chemical_space: ChemicalSpaceBaseLayer,
    n_components: int = 2,
    method: PROJECTION_METHODS = "tsne",
    metric: str = "jaccard",
    seed: int = SEED,
    n_jobs: int = 1,
    **kwargs,
) -> NDArray[np.float_]:
    """
    Project the chemical space data to a lower-dimensional space.

    Args:
        chemical_space (ChemicalSpaceBaseLayer): The chemical space layer to project.
        n_components (int, optional):
            The number of components in the lower-dimensional space. Defaults to 2.
        method (str, optional): The projection method to use. Defaults to "tsne".
        metric (str, optional): The metric to use for the projection.
            Defaults to "jaccard".
        seed (int, optional): The random seed for reproducibility. Defaults to SEED.
        n_jobs (int, optional): The number of parallel jobs to run. Defaults to 1.
        **kwargs: Additional keyword arguments to pass to the projection algorithm.

    Returns:
        NDArray[float]: An array of the projected data
            in the lower-dimensional space.

    """
    if method == "umap":
        from umap import UMAP

        projector = UMAP(
            n_components=n_components,
            random_state=seed,
            metric=metric,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif method == "tsne":
        from sklearn.manifold import TSNE

        projector = TSNE(
            n_components=n_components,
            random_state=seed,
            metric=metric,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif method == "pca":
        from sklearn.decomposition import PCA

        projector = PCA(n_components=n_components, random_state=seed, **kwargs)
    else:
        raise ValueError(f"Invalid projection method: {method}")

    proj = np.array(projector.fit_transform(chemical_space.features))

    return proj


class ChemicalSpaceProjectionLayer(ChemicalSpaceBaseLayer):
    """
    A layer for projecting chemical space data to a lower-dimensional space.
    """

    @lru_cache
    def project(
        self,
        n_components: int = 2,
        method: PROJECTION_METHODS = "tsne",
        seed: int = 42,
        n_jobs: int = -1,
        **method_kwargs,
    ) -> NDArray[np.float_]:
        """
        Project the chemical space data to a lower-dimensional space.

        Args:
            n_components (int, optional): The number of components in
                the lower-dimensional space. Defaults to 2.
            method (str, optional): The projection method to use. Defaults to "tsne".
            seed (int, optional): The random seed for reproducibility. Defaults to 42.
            n_jobs (int, optional): The number of parallel jobs to run. Defaults to -1.
            **method_kwargs: Additional keyword arguments to pass
                to the projection algorithm.

        Returns:
            NDArray[float]: The projeciton of the space

        """
        return project_space(
            self,
            n_components,
            method,
            metric=self.metric,
            seed=seed,
            n_jobs=n_jobs,
            **method_kwargs,
        )



import numpy as np
from numpy.typing import NDArray
from rdkit.Chem import Mol  # type: ignore
from sklearn.neighbors import BallTree

from .base import ChemicalSpaceBaseLayer, T


def _find_overlap(
    cs1: ChemicalSpaceBaseLayer,
    cs2: ChemicalSpaceBaseLayer,
    radius: float = 0.4,
    min_neighbors: int = 1,
    metric: str = "jaccard",
) -> NDArray[np.int_]:
    """
    Find the indices of points in `cs2` that have at least `min_neighbors`
    neighbors in `cs1` within a given `radius` distance.

    Parameters:
        cs1 (ChemicalSpaceBaseLayer): The first chemical space layer.
        cs2 (ChemicalSpaceBaseLayer): The second chemical space layer.
        radius (float, optional): The radius within which to search for neighbors.
            Defaults to 0.4.
        min_neighbors (int, optional): The minimum number of neighbors required
            for a point in `cs2` to be considered overlapping with `cs1`. Defaults to 1.
        metric (str, optional): The metric to use for the BallTree.
            Defaults to "jaccard".

    Returns:
        NDArray[int]: The indices of points in `cs2` that have at least
            `min_neighbors` neighbors in `cs1`.
    """

    tree = BallTree(cs1.features, metric=metric)
    # Array of neighbor counts for each point in cs2
    num_neighbors: NDArray[np.int_] = np.array(
        tree.query_radius(
            cs2.features, r=radius, return_distance=False, count_only=True
        )
    )

    # Indices of points in cs2 that have at least `min_neighbors` neighbors in cs1
    idx_with_neighbors = np.where(num_neighbors >= min_neighbors)[0]

    return idx_with_neighbors


class ChemicalSpaceNeighborsLayer(ChemicalSpaceBaseLayer):
    def find_neighbors(
        self, mol: Mol, radius: float = 0.6, min_neighbors: int = 1  # type: ignore
    ):
        """
        Find the indices of points in `self` that have at least
        `min_neighbors` neighbors within a given `radius` similarity threshold to a
        given molecule.

        Args:
            mol (Mol): The molecule to find neighbors for.
            radius (float, optional): The radius within which to search for neighbors.
                Defaults to 0.6.
            min_neighbors (int, optional):
                The minimum number of neighbors required for a point

        Returns:
            NDArray[int]: The indices of points in `self` that have at least
                `min_neighbors` neighbors.

        """
        other = ChemicalSpaceBaseLayer(mols=(mol,))
        return _find_overlap(other, self, radius, min_neighbors, metric=self.metric)

    def find_overlap(
        self, other: T, radius: float = 0.6, min_neighbors: int = 1  # type: ignore
    ) -> NDArray[np.int_]:
        """
        Find the indices of points in `self` that have at least `min_neighbors`
        neighbors in `other` within a given `radius` similarity threshold.

        Args:
            other (ChemicalSpaceBaseLayer): The other chemical space layer.
            radius (float, optional): The radius within which to search for neighbors.
                Defaults to 0.6.
            min_neighbors (int, optional): The minimum number of neighbors required
                for a point in `other`

        Returns:
            NDArray[int]: The indices of points in `self` that have at least
                `min_neighbors` neighbors in `other`.

        """
        return _find_overlap(other, self, radius, min_neighbors, metric=self.metric)

    def carve(self, other: T, radius: float = 0.6, min_neighbors: int = 1) -> T:
        """
        Remove points from `self` that have at least `min_neighbors` neighbors in
        `other` within a given `radius` similarity threshold.

        Args:
            other (ChemicalSpaceBaseLayer): The other chemical space layer.
            radius (float, optional): The radius within which to search for neighbors.
                Defaults to 0.6.
            min_neighbors (int, optional): The minimum number of neighbors required
                for a point in `other`

        Returns:
            T: A new chemical space layer with points removed.

        """
        idx = self.find_overlap(other, radius, min_neighbors=min_neighbors)
        mask = np.ones(len(self.features), dtype=bool)
        mask[idx] = False
        return self[mask]
